/*
 *	Filename:		readargs.c
 *	Description:	Functions for parsing and extracting command line parameters.
 *					The basic operation of the functions are that 
 *
 *					Define the ALLOW_ESCAPE_QUOTES to the numerical value 1 to allow
 *					escaped double quote marks in the argument values. Might play
 *					havoc with directory parameters, i.e. out_dir="c:\some space dir name\"
 *					Now, do you consider that to be a escaped double quote or a
 *					directory specifier. I wish Windows had a decent command line parser.
 *	To do:			
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "readargs.h"
#include "..\consoles\clutils.h"
#include "..\debug.h"


/* Define this to 1 to allow escaped double quote marks in the argument parameters.
 * See the main file description (above) for explanation.
 */
#ifndef ALLOW_ESCAPE_QUOTES
#define ALLOW_ESCAPE_QUOTES	0
#endif

#define WHITESPACE				" \x09\x0A\x0D"		/* All whitespace characters in a string */
#define TEMPLATE_SEPARATOR		","					/* The string/characters used to separate argument specifiers in a template */
#define TEMPLATE_TYPE			"/"					/* Character used to escape the argument type specifier */
#define TEMPLATE_ALTERNATIVE	"="					/* Character used to separate alternative forms for argument names */

/* Strings containing the different characters for argument type specifiers */
const char * const ta_types[] = {TATS_STRING,
								TATS_NUMBER,
								TATS_BOOLEAN,
								TATS_TOGGLE,
								TATS_KEYWORD,
								TATS_REQUIRED,
								TATS_MULTIPLE,
								TATS_EOL};
								
#define EXE_NAME_LENGTH			512		/* Size of buffer to use when getting executable filename */
#define PROMPT_STRING_LENGTH	512		/* Size of buffer to use when reading user's prompted input from command line */

#define MAXIMUM_ARGUMENTS		100		/* 100 command line arguments can be parsed */


/* Structure used when parsing the command line to store and associate argument names and values */
struct CommandLineArgument
{
	char	*cla_Argument;		/* Pointer to the string (if specified) containing the argument name */
	char	*cla_Value;			/* Pointer to the string (if specified) containing the argument value */
};


/* Internal function definitions */
static long				ra_CountStrings(char *string, char *delimiter);
static char				*ra_mystrtok(char *s, char *ct, char replace);
static long				ra_ParseTemplate(char *template_string, struct Arguments *args);
static long				ra_ParseCommandLine(struct Arguments *args, LongPtrType *arg_values, char *template_string, char *alternative);
static unsigned long	ra_ArgumentType(struct Arguments *args, long arg_number);
static void				ra_SeparateAlternatives(struct Arguments *args, long arg_number);
static struct TemplateArgument	*ra_FindArgumentName(struct Arguments *args, struct TemplateArgument *start_from, char *name);
static struct TemplateArgument	*ra_FindArgumentType(struct Arguments *args, struct TemplateArgument *start_from, unsigned long mask, unsigned long pattern);
static long				ra_ScanForPrompt(struct Arguments *args, char *template_string, char *command_line);
static long				ra_SkipCommandName(char *command_string);
static inline long		ra_nbits(unsigned long x);
static inline long		ra_log2(unsigned long x);


/*
 *	Function:		ReadArguments
 *	Synopsis:		args = ReadArguments(template_string, arg_values, reserved)
 *	Parameters:		char *template_string		- The template string containing the argument names, formats and types (see below).
 *					LongPtrType *arg_values		- Pointer to an array of sap's which contain the filled in
 *												  results of the argument parsing. This should contain the default values
 *												  before calling this function (see below).
 *					struct Arguments *reserved	- Reserved, set to NULL for now/ever.
 *	Returns:		struct Arguments *args - Pointer to a newly allocated base structure or NULL for (allocation) failure.
 *	Description:	Parses the template string and command line arguments for a Win32
 *					program and extracts any values specified on the command line
 *					into a user's array. If there is an error during parsing the
 *					template or command line then the arg_Error field will be set
 *					to something other than ARG_ERROR_NONE.
 *
 *					A call to this function MUST be matched by a call to FreeArguments,
 *					unless this function fails with a return value of NULL (although
 *					passing a NULL value to FreeArguments is safe).
 *
 *					The format of the template string is as follows (angle brackets
 *					mean to replace them with the thing described inside them, square
 *					brackets denote something optional and "..." means a continuation
 *					of the previous thing):
 *						"<argument_name1>[=<alternative name1.1>[=<alternative name1.2>[=<alternative name1.3>...]]][/<type>[/<type>][...],<argument_name2>[=<alternative name2.1>[=<alternative name2.2>[=<alternative name2.3>...]]][/<type>[/<type>][,...]"
 *
 *					Note that the real argument name must be the first alternative
 *					specified and that all type modifiers MUST be at the end of the
 *					argument.
 *
 *					On the command line argument names are case insensitive. Quotes
 *					can be used to enclose a string with spaces or other special
 *					characters and have it treated as a single string. The question
 *					mark on a command line, outwith quotes signifies that the user
 *					wants to be prompted with the template to enter parts of the
 *					command line.
 *
 *					See also the descriptions for ra_ParseTemplate() and ra_ParseCommandLine()
 *					for some important points.
 */
struct Arguments	*ReadArguments(char *template_string, LongPtrType *arg_values, char *alternative_arguments)
{
	struct Arguments	*parsed_args = NULL;	/* The base data structure for the template and argument parsing */
	
	/* Make sure the pointers sent to this function are valid */
	if(template_string && arg_values)
	{
		/* Allocate some memory for the base data structure */
		parsed_args = calloc(1, sizeof(struct Arguments));
		if(NULL != parsed_args)
		{
			/* Parse the template */
			if(ARG_ERROR_NONE == ra_ParseTemplate(template_string, parsed_args))
			{
				/* If that was successful, parse the command line */
				ra_ParseCommandLine(parsed_args, arg_values, template_string, alternative_arguments);
			}
		}
	}
	
	return(parsed_args);
}


/*
 *	Function:		FreeArguments
 *	Synopsis:		FreeArguments(args)
 *	Parameters:		struct Arguments *args - Pointer to an Arguments structure, as returned by ReadArguments
 *	Returns:		Nothing
 *	Description:	Frees all memory associated with the parsing of the arguments.
 *					This must be called for any call to ReadArguments, and although it
 *					is not strictly necessary if ReagArguments returns NULL this
 *					function will handle that case correctly.
 */
void				FreeArguments(struct Arguments *args)
{
	int		argitem;		/* Used as a loop counter when freeing arrays of argument items */
	
	/* Make sure we have a valid pointer to the base data */
	if(NULL != args)
	{
		/* Free the copy of the template */
		if(NULL != args->arg_Template)
		{
			free(args->arg_Template);
			args->arg_Template = NULL;
		}
		
		/* Free the copy of the command line */
		if(NULL != args->arg_CommandLine)
		{
			free(args->arg_CommandLine);
			args->arg_CommandLine = NULL;
		}
		
		/* Free each item in the parsed template list - all template argument names
		 * and command line argument strings point to somewhere in the copy of the
 		 * template or command line, therefore we do not need to free them.
   		 */
	 	if(args->arg_Arguments)
	 	{
	 		/* For each argument type */
			for(argitem = 0; argitem < args->arg_NumberArgs; argitem++)
			{
				/* Check if there are any alternatives allocated */
				if(args->arg_Arguments[argitem].ta_Alternatives)
				{
					/* And if so, free the alternatives array */
					free(args->arg_Arguments[argitem].ta_Alternatives);
					args->arg_Arguments[argitem].ta_Alternatives = NULL;
				}
				
				/* If there is a multiple type of argument and we have allocated the array, free that */
				if((args->arg_Arguments[argitem].ta_Type & TATF_MULTIPLE) && NULL != args->arg_Arguments[argitem].ta_CommandLineArg)
				{
					free(args->arg_Arguments[argitem].ta_CommandLineArg);
					args->arg_Arguments[argitem].ta_CommandLineArg = NULL;
				}
			}
			
			/* Free the arguments array */
			free(args->arg_Arguments);
			args->arg_Arguments = NULL;
		}
		
		/* Free the base data structure */
		free(args);
		args = NULL;	/* Unnecessary, but nice to do if code ever appears under here (will help in debugging if we get accesses to NULL pointers) */
	}
}


/*
 *	Function:		FindArgumentByNumber
 *	Synopsis:		argument_name = FindArgumentByNumber(template_string, argument_number, buffer, buffer_length)
 *	Parameters:		char *template_string	- Pointer to a string containing the template in which to search for an argument name
 *					int argument_number		- The number of the argument (position in the template string) to search for
 *					char *buffer			- Pointer to the buffer in which to store the argument name if found
 *					int buffer_length		- Length of buffer
 *	Returns:		char *argument_name		- Pointer to the start of the buffer if the argument name was found, otherwise NULL.
 *	Description:	Searches through a ReadArgs() style template string for the
 *					argument name at the specified position. If the argument name
 *					is found then it will be copied into the buffer and null terminated.
 *					If the buffer is too small then as much of the string will be
 *					copied as possible (still null terminated) but the return
 *					value will be NULL. If the argument cannot be found then the
 *					buffer will be set to an empty string and the return value
 *					will be NULL.
 */
char	*FindArgumentByNumber(char *template_string, int argument_number, char *buffer, unsigned long buffer_length)
{
	char				*argument_name = NULL;		/* Pointer to argument name if found or NULL */
	struct Arguments	*parsed_args = NULL;		/* The base data structure for the template and argument parsing */

	if(NULL != template_string && NULL != buffer && argument_number >= 0)
	{
		/* Initially we have no argument name */
		*buffer = '\0';
		
		/* Allocate some memory for the base data structure */
		parsed_args = calloc(1, sizeof(struct Arguments));
		if(NULL != parsed_args)
		{
			/* Parse the template */
			if(ARG_ERROR_NONE == ra_ParseTemplate(template_string, parsed_args))
			{
				/* Check pointers and number of arguments within parsed template */
				if(argument_number < parsed_args->arg_NumberArgs && NULL != parsed_args->arg_Arguments)
				{
					if(NULL != parsed_args->arg_Arguments[argument_number].ta_ArgumentName)
					{
						/* If buffer size is big enough to take entire argument name then we return
						 * the pointer to the buffer.
						 */
						if(strlen(parsed_args->arg_Arguments[argument_number].ta_ArgumentName) <= (buffer_length - 1))
						{
							argument_name = buffer;
						}
						
						/* Copy the argument name into the buffer, up to the size of the buffer and leave space for terminator */
						strncpy(buffer, parsed_args->arg_Arguments[argument_number].ta_ArgumentName, buffer_length - 1);
						
						/* Do I need to manually terminate the string if it copied the full size? */
					}
				}
			}
			
			FreeArguments(parsed_args);
		}
	}

	return(argument_name);
}


/*
 *	Function:		ra_CountStrings
 *	Synopsis:		number_strings = ra_CountStrings(string, delimiter)
 *	Parameters:		char *string	- Address of the start of a character string in which to count occurrances of sub strings
 *					char *delimiter	- The delimiter used to separate the sub strings
 *	Returns:		long number_strings - The number of sub strings found in the string, separated by the delimiter
 *	Description:	Counts the number of substrings within a string, each one separated by
 *					some delimiter string. Kind of similar to a lot of the standard string library,
 *					but none do this.
 */
static long	ra_CountStrings(char *string, char *delimiter)
{
	long	number_strings = 0;		/* Count of the number of strings separated by the delimiter */
	int		char_position;			/* Position in the string while parsing it */
	char	*found_char;			/* Pointer to character when searching in strings */
	
	/* Make sure we have a string pointer and the delimiter is not the null character
	 * (since we wouldn't be able to distinguish between that and the end of the string -
	 * possibly change for future to be able to specify the length of the string instead
	 * of relying on null terminators, since that could be quite useful
	 */
	if(NULL != string && NULL != delimiter && '\0' != delimiter[0])
	{
		/* Start at the start of the string and repeat until we run out of delimiters */
		char_position = 0;
		do
		{
			d1(printf(__FILE__ "/%s/%d: Counting strings, character position=%d\n", __func__, __LINE__, char_position));
			
			/* Search for a delimiter in the string */
			found_char = strstr(&string[char_position], delimiter);
			if(NULL != found_char)
			{
				/* Skip past any whitespace and ignore empty argument type definitions */
				char_position += strspn(&string[char_position], WHITESPACE);
				if(string[char_position] != ',')
				{
					number_strings++;
				}

				/* Move onto character after comma in string */
				char_position += (found_char - &string[char_position]) + 1;
			}
		} while(found_char);
				
		/* Any arguments defined after the last comma? */
		if((strlen(&string[char_position]) - strspn(&string[char_position], WHITESPACE)) > 0)
		{
			number_strings++;
		}
	}
	
	return(number_strings);
}


/*
 *	Function:		ra_mystrtok
 *	Synopsis:		token_position = ra_mystrtok(string, tokens, replace)
 *	Parameters:		char *string  - Pointer to the string which to separate into
 *									substrings.
 *					char *tokens  - Pointer to the string which contains valid
 *									separator characters at which to split the string.
 *					char replace  - Character to replace the found token with
 *	Returns:		char *token_position - Pointer to the 
 *	Description:	Like the standard C string library strtok function but works
 *					in a way that is more useful for separating strings based on
 *					some token, rather than mangling surrounding characters and
 *					requiring massive strings for the token/ct list.
 *
 *					This function searches a string for a token character and
 *					if found replaces the token character with a replacement character
 *					and returns a pointer to this replacement character's position
 *					in the string.
 *
 *					The first time you call this function you should specify a
 *					valid string pointer. Each successive call to parse the remaining
 *					contents of the same string should be done with a NULL pointer
 *					for the string parameter.
 *
 *					The tokens parameter can change on each call to this function.
 *					Each character in the tokens string is a valid separator, not the
 *					string as a whole.
 *
 *					This function is not thread safe since it uses static data storage.
 *	Example:
 *	{
 *		char	parse_me[] = "asdfghjk,,,   ,xcvbnm,ertyui,fghjk";
 *		char	*tok_pos = parse_me;
 *		char	*start_str = parse_me;
 *		
 *		tok_pos = ra_mystrtok(tok_pos, ",", '\0');
 *		printf("Extracted string: <%s>\n", start_str);
 *		while(tok_pos)
 *		{
 *			start_str = tok_pos + 1;
 *			tok_pos = ra_mystrtok(NULL, ",", '\0');
 *			printf("Extracted string: <%s>\n", start_str);
 *		}
 *	}
 */
static char	*ra_mystrtok(char *string, char *tokens, char replace)
{
	static char	*current_position = NULL;
	char		*token_position = NULL;
	
	if(string)
	{
		current_position = string;
	}
	
	/* If we have a (hopefully) valid string pointer */
	if(current_position)
	{
		/* Find next token character or end of string */
		current_position += strcspn(current_position, tokens);
		
		/* If it is not the terminating null, replace with the replacement character
		 * and keep pointer to replaced character.
		 */
		if(*current_position != '\0')
		{
			token_position = current_position;
			*current_position++ = replace;
		}
		else
		{
			/* Since we reached the end of the string, we return NULL */
			current_position = token_position = NULL;
		}
	}
	
	return(token_position);
}


/*
 *	Function:		ra_ParseTemplate
 *	Synopsis:		error = ra_ParseTemplate(template_string, args)
 *	Parameters:		char *template_string	-
 *					struct Arguments *args	-
 *	Returns:		long error - The error code showing what was wrong (if anything) with the
 *								 parsing of the template string.
 *	Description:	Parses the template that is to be used to extract parameters from the command line.
 *					Performs all operations required to get the template into the
 *					correct format, including all memory allocations, string copies
 *					and separations.
 */
static long	ra_ParseTemplate(char *template_string, struct Arguments *args)
{
	char	*found_char;					/* Pointer to character when searching in strings */
	char	*current_char;					/* Pointer to the current character */
	char	*end_of_template;				/* Pointer to null terminator in template, used as a loop validation */
	int		argument;						/* Counter for the current argument while separating them */
	long	error = ARG_ERROR_INVALID_ARGS;	/* Default return code, representing invalid parameters to this command */
	char	*token_position;				/* Pointer to token character when separating argument strings */
	int		has_multiple = 0;				/* Flag showing whether a multiple argument type has been found */


	/* Make sure we have a valid base data structure pointer */
	if(NULL != args)
	{
		/* Allocate memory and take working copy of the template.
		 * This will be butchered into our own internal format (basically have NULLs
		 * scattered throughout.
		 */
		args->arg_Template = calloc(strlen(template_string) + 1, sizeof(char));
	
		/* Check that it was allocated successfully */			
		if(NULL != args->arg_Template)
		{
			strcpy(args->arg_Template, template_string);			/* Copy user template */
									
			/* Count number of arguments */
			args->arg_NumberArgs = ra_CountStrings(args->arg_Template, TEMPLATE_SEPARATOR);
			d1(printf(__FILE__ "/%s/%d: Found %ld arguments\n", __func__, __LINE__, args->arg_NumberArgs));
			
			/* If we have some arguments to recognise, allocate an array to store their names
			 * and information
			 */
			if(args->arg_NumberArgs > 0)
			{
				args->arg_Arguments = calloc(args->arg_NumberArgs, sizeof(struct TemplateArgument));
				if(NULL != args->arg_Arguments)
				{
					/* Tokenise the template string by replacing the commas with
					 * nulls and keeping copies of the pointers to the starts of each
					 * string.
					 */
	 				/* Starting with the first argument */
	 				argument = 0;
	 				
	 				/* Find first comma and replace with NULL to terminate the argument name string */
	 				args->arg_Arguments[argument].ta_ArgumentName = token_position = args->arg_Template;
	 				token_position = ra_mystrtok(token_position, TEMPLATE_SEPARATOR, '\0');
	 				
	 				/* Skip all whitespace at start of string, and move onto the next argument if
	 				 * we actually have an argument specifier string with something in it.
	 				 */
	 				args->arg_Arguments[argument].ta_ArgumentName += strspn(args->arg_Arguments[argument].ta_ArgumentName, WHITESPACE);
	 				if(strlen(args->arg_Arguments[argument].ta_ArgumentName) > 0)
	 				{
	 					argument++;
	 				}
	 				
	 				while(argument < args->arg_NumberArgs && NULL != token_position)
	 				{
		 				/* Find first comma and replace with NULL to terminate the argument name string */
		 				args->arg_Arguments[argument].ta_ArgumentName = token_position + 1;
		 				token_position = ra_mystrtok(NULL, TEMPLATE_SEPARATOR, '\0');

		 				/* Skip all whitespace at start of string, and move onto the next argument if
		 				 * we actually have an argument specifier string with something in it.
		 				 */
		 				args->arg_Arguments[argument].ta_ArgumentName += strspn(args->arg_Arguments[argument].ta_ArgumentName, WHITESPACE);
		 				if(strlen(args->arg_Arguments[argument].ta_ArgumentName) > 0)
		 				{
		 					argument++;
		 				}
	 				}
	 				
	 				/* Check that we managed to extract all argument specifier strings */
	 				if(argument == args->arg_NumberArgs)
	 				{
						for(argument = 0; argument < args->arg_NumberArgs && args->arg_Error == ARG_ERROR_NONE; argument++)
						{
		 					d1(printf(__FILE__ "/%s/%d: Separated argument string <%s>\n", __func__, __LINE__, args->arg_Arguments[argument].ta_ArgumentName));
							
							/* Search for argument type modifiers, replacing with whitespace? */
							args->arg_Arguments[argument].ta_Type = ra_ArgumentType(args, argument);
							d1(printf(__FILE__ "/%s/%d: names=<%s>  type=%ld  error=%d\n", __func__, __LINE__, args->arg_Arguments[argument].ta_ArgumentName, args->arg_Arguments[argument].ta_Type, args->arg_Error));
							
							/* Check that one one multiple argument type has been specified in the template */
							if(args->arg_Arguments[argument].ta_Type & TATF_MULTIPLE)
							{
								if(has_multiple)
								{
									args->arg_Error = ARG_ERROR_MULTIPLES;
									args->arg_ErrorPosition = argument;
								}
								else
								{
									has_multiple = 1;
								}
							}
							
							if(ARG_ERROR_NONE == args->arg_Error)
							{
								/* Count number of alternatives */
								args->arg_Arguments[argument].ta_NumberAlternatives = ra_CountStrings(args->arg_Arguments[argument].ta_ArgumentName, TEMPLATE_ALTERNATIVE) - 1;
								d1(printf(__FILE__ "/%s/%d: alternatives=%ld\n", __func__, __LINE__, args->arg_Arguments[argument].ta_NumberAlternatives));
								if(args->arg_Arguments[argument].ta_NumberAlternatives > 0)
								{
									/* Allocate array for them */
									args->arg_Arguments[argument].ta_Alternatives = calloc(args->arg_Arguments[argument].ta_NumberAlternatives, sizeof(char *));
									if(NULL != args->arg_Arguments[argument].ta_Alternatives)
									{
										d1(int		d2_i);		/* Used by the debug output below */
										
										/* Parse and record starts of alternative forms, ignoring whitespace and/or replacing with nulls */
										ra_SeparateAlternatives(args, argument);
										
										/* Just some debug output for checking */
										d1(printf(__FILE__ "/%s/%d: Main name is <%s>\n", __func__, __LINE__, args->arg_Arguments[argument].ta_ArgumentName);
          								for(d2_i = 0; d2_i < args->arg_Arguments[argument].ta_NumberAlternatives; d2_i++)
										{
											printf(__FILE__ "/%s/%d: Alternative %d is <%s>\n", __func__, __LINE__, d2_i, args->arg_Arguments[argument].ta_Alternatives[d2_i]);
										})
									}
									else
									{
										/* Failed to allocate memory for the alternatives. We could still proceed
										 * to use the argument as it is, but if it was a keyword type or the user
										 * wanted to type it in, they would need to use the full form with all alternative
										 * names. So, not much fun or use.
										 */
					 					args->arg_Error = ARG_ERROR_ALLOCATION;
	 									args->arg_ErrorPosition = argument;
									}
								}
							}
						}
	 				}
	 				else
	 				{
	 					/* If we could not extract all strings, return an error to show this and
	 					 * let the calling function know where we failed.
	 					 */
	 					args->arg_Error = ARG_ERROR_ARGUMENT_SEPARATION;
	 					args->arg_ErrorPosition = argument;
	 				} /* Number of separated arguments */
				}
				else
				{
					/* Failed to allocate array for arguments to recognise */
					args->arg_Error = ARG_ERROR_TEMPLATE_ALLOCATION;
					args->arg_ErrorPosition = -1;
				} /* args->arg_Arguments array allocated */
			} /* args->arg_NumberArgs > 0 */
		}
		else
		{
			/* Failed to allocate memory for the copy of the template string */
			args->arg_Error = ARG_ERROR_ALLOCATION;
			args->arg_ErrorPosition = -1;
		} /* args->arg_Teamplate allocated */
		
		/* Get the real error return value from the args structure */
		error = args->arg_Error;
	} /* args valid */
	
	return(error);
}


/*
 *	Function:		ra_CountEscapes
 *	Synopsis:		number_escapes = ra_CountEscapes(quote_position, start)
 *	Parameters:		char *quote_position	- Position of a double quote character in a string
 *					char *start				- 
 *	Returns:		
 *	Description:	
 */
static int	ra_CountEscapes(char *quote_position, char *start)
{
	int		number_escapes = 0;
	
	quote_position--;
	
	while('\\' == *quote_position && (unsigned long)quote_position >= (unsigned long)start)
	{
		number_escapes++;
		quote_position--;
	}
	
	return(number_escapes);
}


/*
 *	Function:		ra_ParseCommandLine
 *	Synopsis:		error = ra_ParseCommandLine(args)
 *	Parameters:		struct Arguments *args	- Address of a valid base data structure which has had its
 *											  template parsed correctly and should now have the command
 *											  line parsed through it
 *					LongPtrType *arg_values	- 
 *	Returns:		long error - The status of any error which occurs during the
 *								 parsing of the command line
 *	Description:	
 *
 *					You must provide a valid stdin before calling this function
 *					if one does not already exist (i.e. a Windows GUI program)
 *					because if the question mark symbol is found on the command
 *					line, not enclosed by quotes, it is taken as meaning to prompt
 *					the user. The reason stdin is required is that this function
 *					will print the prompt to stdin and read it's input from stdin,
 *					allowing the user to interact with the program even when the
 *					output is being redirected.
 *
 *					[[<symbols>]<argument name>[ws][=]][ws]["][<argument value>]["]
 */
static long	ra_ParseCommandLine(struct Arguments *args, LongPtrType *arg_values, char *template_string, char *alternative)
{
	long						error = ARG_ERROR_INVALID_ARGS;	/* Default return code, representing invalid parameters to this command */
	char						*cmd_position;					/* Pointer to current position while parsing command line */
	char						*arg_start;
	char						*arg_end;
	int							end_of_string;
	struct CommandLineArgument	split_args[MAXIMUM_ARGUMENTS];
	int							current_split;
	int							shuffle_split;							/* Used when removing a multiple value from the front of the array and we need to shuffle all the others up by one */
	
	int							previous_extracted_type = TATB_NONE;
	int							current_extracted_type = TATB_NONE;
	struct TemplateArgument		*multiple;
	struct TemplateArgument		*current_argument;
	char						terminator;
	
	long						argument_number;
	int							found_multiple;

	d1(printf(__FILE__ "/%s/%d: Entered\n", __func__, __LINE__));

	/* Make sure we have a valid base data structure pointer */
	if(NULL != args)
	{
		/* Allocate a zeroed buffer for the number of characters in the command line string */
		if(NULL != alternative)
		{
			args->arg_CommandLine = calloc(strlen(alternative) + 1, sizeof(char));
		}
		else
		{
			args->arg_CommandLine = calloc(lstrlen(GetCommandLine()) + 1, sizeof(char));
		}
		
		if(NULL != args->arg_CommandLine)
		{
			/* If successful, take an ASCII/OEM/8bit copy of it (since it is a Unicode string) */
			if(NULL == alternative)
			{
				CharToOem(GetCommandLine(), args->arg_CommandLine);
				d1(printf(__FILE__ "/%s/%d: OEM string=<%s>\n", __func__, __LINE__, args->arg_CommandLine));
			
				/* Get pointer to start of command line for parsing */
				cmd_position = args->arg_CommandLine + ra_SkipCommandName(args->arg_CommandLine);
			}
			else
			{
				strcpy(args->arg_CommandLine, alternative);
				cmd_position = args->arg_CommandLine;
			}
			
			/* Just chicken... */
			d1(printf(__FILE__ "/%s/%d: Command line to be parsed=<%s>\n", __func__, __LINE__, cmd_position));

			ra_ScanForPrompt(args, template_string, cmd_position);
			d1(printf(__FILE__ "/%s/%d: Full command line=<%s>\n", __func__, __LINE__, args->arg_CommandLine));

			if(NULL == alternative)
			{
				/* Get pointer to start of command line for parsing, since it might have changed */
				cmd_position = args->arg_CommandLine + ra_SkipCommandName(args->arg_CommandLine);
				d1(printf(__FILE__ "/%s/%d: Full parse without command=<%s>\n", __func__, __LINE__, cmd_position));
			}

			/* Find the multiple type on the command line, if there is one */
			multiple = current_argument = ra_FindArgumentType(args, NULL, TATF_MULTIPLE, TATF_MULTIPLE);

			end_of_string = 0;
			arg_end = cmd_position;
			current_split = 0;
			do
			{
				arg_start = arg_end + strspn(arg_end, WHITESPACE);
				if('\0' != *arg_start)
				{
					if('\"' == *arg_start)
					{
						/* The loop structure here permits \" style quotes to be included in the
						 * parameter and not treated as the end of the parameter. If you need to
						 * be looking for \" at the end of a string then you will need to get your
						 * users to use \\" and translate it into a single \ yourself.
						 */
		 				#if ALLOW_ESCAPE_QUOTES==1
						do
						#endif
						{
							arg_start++;
							arg_end = arg_start + strcspn(arg_start, "\"");
						}
		 				#if ALLOW_ESCAPE_QUOTES==1
						while((ra_CountEscapes(arg_start, arg_end) & 1) == 1);
						#endif
							
            			terminator = *arg_end;
						*arg_end = '\0';
						current_extracted_type = TATB_VALUE;
						
						d1(printf(__FILE__ "/%s/%d: Found quoted value=<%s>\n", __func__, __LINE__, arg_start));
					}
					else if('=' == *arg_start)
					{
						arg_end = (arg_start + 1) + strcspn((arg_start + 1), "=");
            			terminator = *arg_end;
						current_extracted_type = TATB_NONE;
					}
					else
					{
            			arg_end = arg_start + strcspn(arg_start, WHITESPACE "=");
            			terminator = *arg_end;
            			*arg_end = '\0';
            			
            			d1(printf(__FILE__ "/%s/%d: Found string=<%s>\n", __func__, __LINE__, arg_start));
            			
            			if(previous_extracted_type != TATB_STRING && previous_extracted_type != TATB_NUMBER)
            			{
            				d1(printf(__FILE__ "/%s/%d: Did not just extract a string or number argument\n", __func__, __LINE__));
            				
            				current_argument = ra_FindArgumentName(args, NULL, arg_start);
            				if(NULL != current_argument)
            				{
            					d1(printf(__FILE__ "/%s/%d: Matched argument name\n", __func__, __LINE__));
            					
            					current_extracted_type = ra_log2(current_argument->ta_Type & (TATF_STRING|TATF_NUMBER|TATF_BOOLEAN|TATF_TOGGLE));
            					d1(printf(__FILE__ "/%s/%d: Extracted type is %d\n", __func__, __LINE__, current_extracted_type));
            					
            					if(current_argument->ta_Type & TATF_EOL)
            					{
            						if(terminator != '\0')
            						{
            							current_argument->ta_CommandLineArg = arg_end + 1;
        							}
        							else
        							{
        								current_argument->ta_CommandLineArg = arg_end;
    								}
            						end_of_string = 1;
            					}
            				}
            				else
        					{
        						d1(printf(__FILE__ "/%s/%d: Did not match argument name\n", __func__, __LINE__));
        						
        						current_argument = multiple;
        						current_extracted_type = TATB_VALUE;
    						}
						}
						else
						{
							current_extracted_type = TATB_VALUE;
						}
					}

					if(current_extracted_type != TATB_NONE)
					{
						if(current_extracted_type == TATB_VALUE)
						{
							if(NULL != current_argument)
							{
								if(current_argument == multiple)
								{
									split_args[current_split++].cla_Value = arg_start;
									/* Check MAXIMUM_ARGUMENTS */
								}
								else
								{
									current_argument->ta_CommandLineArg = arg_start;
									current_argument = multiple;
								}
							}
							else
							{
								/* Discard - do nothing? */
							}
						} /* current extracted type is value */
						else
						{
							/* Extracted an argument */
							if(current_extracted_type != TATB_STRING && current_extracted_type != TATB_NUMBER)
							{
								/* Not a string or number, so just store itself in the command line for the argument */
								current_argument->ta_CommandLineArg = arg_start;
								
								current_argument = multiple;
							}
						} /* current extracted type is argument */
			
      					/* Set previous extracted type to current extracted type */			
						previous_extracted_type = current_extracted_type;
					} /* current_extracted_type != none */

					/* Move past the null we have inevitably stuck into the command line, unless we already found it */
					if(terminator != '\0')
					{
						arg_end++;
					}
				} /* Found character not null */
				else
				{
					end_of_string = 1;
				}
			} while(!end_of_string);


			/* Should probably find out what the feck I've done */
			d1({
				long	carg;
    			long	csplt;
				
				printf(__FILE__ "/%s/%d: Template matches\n", __func__, __LINE__);
				for(carg = 0; carg < args->arg_NumberArgs; carg++)
				{
					printf(__FILE__ "/%s/%d: Arg=<%s> Val=<%s>\n", __func__, __LINE__, args->arg_Arguments[carg].ta_ArgumentName, args->arg_Arguments[carg].ta_CommandLineArg == NULL ? "NULL" : args->arg_Arguments[carg].ta_CommandLineArg);
				}
				
				printf(__FILE__ "/%s/%d: Remaining matches\n", __func__, __LINE__);
				for(csplt = 0; csplt < current_split; csplt++)
				{
					printf(__FILE__ "/%s/%d: Split %ld Val=<%s>\n", __func__, __LINE__, csplt, split_args[csplt].cla_Value);
				}
   			});
   			
   			/* Go from the back to the front, stopping at the first multiple argument so we can steal
   			 * values from the end of it to use with unfulfilled required/non-keyword arguments.
   			 */
   			for(found_multiple = 0, argument_number = args->arg_NumberArgs; argument_number && ARG_ERROR_NONE == args->arg_Error && !found_multiple; )
   			{
   				argument_number--;
   				d1(printf(__FILE__ "/%s/%d: Checking argument number=%d\n", __func__, __LINE__, argument_number));
   				
				if(args->arg_Arguments[argument_number].ta_Type & TATF_MULTIPLE)
   				{
   					/* Will process the multiple argument when going forwards so we do not want to do that now */
   					d1(printf(__FILE__ "/%s/%d: Going backwards found multiple\n", __func__, __LINE__));
   						
					/* Break out of this loop */
					found_multiple = 1;
				}
				else if(args->arg_Arguments[argument_number].ta_Type & TATF_REQUIRED)
				{
					d1(printf(__FILE__ "/%s/%d: Found required\n", __func__, __LINE__));
					
					if(NULL == args->arg_Arguments[argument_number].ta_CommandLineArg)
					{
						if(args->arg_Arguments[argument_number].ta_Type & TATF_KEYWORD)
						{
							d1(printf(__FILE__ "/%s/%d: Required with keyword - can't rob multiple\n", __func__, __LINE__));
							args->arg_Error = ARG_ERROR_REQUIRED_MISSING;
	   						args->arg_ErrorPosition = argument_number;
						}
						else
						{
							d1(printf(__FILE__ "/%s/%d: Needs its argument  multiple=0x%08lx  current_split=%ld\n", __func__, __LINE__, multiple, current_split));
							
	     					if(NULL != multiple && current_split > 0)
							{
								current_split--;
								args->arg_Arguments[argument_number].ta_CommandLineArg = split_args[current_split].cla_Value;
								d1(printf(__FILE__ "/%s/%d: Split available <%s>\n", __func__, __LINE__, args->arg_Arguments[argument_number].ta_CommandLineArg));
							}
							else
							{
								d1(printf(__FILE__ "/%s/%d: No split available\n", __func__, __LINE__));
								args->arg_Error = ARG_ERROR_REQUIRED_MISSING;
		   						args->arg_ErrorPosition = argument_number;
							}
						}
					}
				}
			}


   			/* Go from the front to the back, stopping at the first multiple argument so we can steal
   			 * values from the start of it to use with unfulfilled required/non-keyword arguments.
   			 * Only need to do this if we found the multiple item, otherwise we will have already
   			 * processed all the items.
   			 */
		 	if(found_multiple)
		 	{
	   			for(found_multiple = 0, argument_number = 0; argument_number < args->arg_NumberArgs && ARG_ERROR_NONE == args->arg_Error && !found_multiple; argument_number++)
				{
	   				d1(printf(__FILE__ "/%s/%d: Checking argument number=%d\n", __func__, __LINE__, argument_number));
   				
	   				if(args->arg_Arguments[argument_number].ta_Type & TATF_MULTIPLE)
	   				{
	   					d1(printf(__FILE__ "/%s/%d: Going forward found multiple\n", __func__, __LINE__));
	   					
	   					if((args->arg_Arguments[argument_number].ta_Type & TATF_REQUIRED) && current_split < 1)
	   					{
	   						args->arg_Error = ARG_ERROR_REQUIRED_MISSING;
	   						args->arg_ErrorPosition = argument_number;
						}
						else
						{
							/* Allocate new storage for the split arguments and copy the
	      					 * pointers to the strings in the split list to the multiple
	            			 * template argument
	            			 */
	            			args->arg_Arguments[argument_number].ta_CommandLineArg = calloc(current_split+1, sizeof(char *));
	            			d1(printf(__FILE__ "/%s/%d: Allocated memory for multiple array = 0x%08lX\n", __func__, __LINE__, args->arg_Arguments[argument_number].ta_CommandLineArg));
	            			
	            			if(NULL != args->arg_Arguments[argument_number].ta_CommandLineArg)
	            			{
	            				while(current_split)
	            				{
	            					current_split--;
	            					
	            					((char **)(args->arg_Arguments[argument_number].ta_CommandLineArg))[current_split] = split_args[current_split].cla_Value;
	        						d1(printf(__FILE__ "/%s/%d: Copied pointer 0x%08lX=<%s> as 0x%08lX=<%s>\n", __func__, __LINE__,
	              						((char **)(args->arg_Arguments[argument_number].ta_CommandLineArg))[current_split],
	              						((char **)(args->arg_Arguments[argument_number].ta_CommandLineArg))[current_split] ? ((char **)(args->arg_Arguments[argument_number].ta_CommandLineArg))[current_split] : "NULL",
	                    				split_args[current_split].cla_Value,
	                    				split_args[current_split].cla_Value ? split_args[current_split].cla_Value : "NULL"));
	            				}
	            			}
						}
						
						/* Break out of this loop */
						found_multiple = 1;
					}
					else if(args->arg_Arguments[argument_number].ta_Type & TATF_REQUIRED)
					{
						d1(printf(__FILE__ "/%s/%d: Found required\n", __func__, __LINE__));
						
						if(NULL == args->arg_Arguments[argument_number].ta_CommandLineArg)
						{
							if(args->arg_Arguments[argument_number].ta_Type & TATF_KEYWORD)
							{
								d1(printf(__FILE__ "/%s/%d: Required with keyword - can't rob multiple\n", __func__, __LINE__));
								args->arg_Error = ARG_ERROR_REQUIRED_MISSING;
		   						args->arg_ErrorPosition = argument_number;
							}
							else
							{
								d1(printf(__FILE__ "/%s/%d: Needs its argument  multiple=0x%08lx  current_split=%ld\n", __func__, __LINE__, multiple, current_split));
								
		     					if(NULL != multiple && current_split > 0)
								{
									args->arg_Arguments[argument_number].ta_CommandLineArg = split_args[0].cla_Value;
									
									/* Shuffle the rest of the split arguments up one array position */
									for(shuffle_split = 1; shuffle_split < current_split; shuffle_split++)
									{
										split_args[shuffle_split - 1] = split_args[shuffle_split];
									}
									
									current_split--;
									d1(printf(__FILE__ "/%s/%d: Split available <%s>\n", __func__, __LINE__, args->arg_Arguments[argument_number].ta_CommandLineArg));
								}
								else
								{
									d1(printf(__FILE__ "/%s/%d: No split available\n", __func__, __LINE__));
									args->arg_Error = ARG_ERROR_REQUIRED_MISSING;
			   						args->arg_ErrorPosition = argument_number;
								}
							}
						}
					}
				}
			}
   
			/* Process all the arguments, putting them into their user array position */
			for(argument_number = 0; argument_number < args->arg_NumberArgs && ARG_ERROR_NONE == args->arg_Error; argument_number++)
   			{
				if(NULL != args->arg_Arguments[argument_number].ta_CommandLineArg)
				{
					d1(printf(__FILE__ "/%s/%d: Checking argument %ld type\n", __func__, __LINE__, argument_number));
					
					switch(args->arg_Arguments[argument_number].ta_Type & (TATF_STRING|TATF_NUMBER|TATF_BOOLEAN|TATF_TOGGLE))
					{
						case TATF_STRING:
							if(args->arg_Arguments[argument_number].ta_Type & TATF_MULTIPLE)
							{
								printf(__FILE__ "/%s/%d: Multiple string = \n", __func__, __LINE__);
        						for(end_of_string = 0; ((char **)args->arg_Arguments[argument_number].ta_CommandLineArg)[end_of_string]; end_of_string++)
        						{
        							printf(__FILE__ "/%s/%d:     0x%08lX     %s\n", __func__, __LINE__, ((char **)args->arg_Arguments[argument_number].ta_CommandLineArg)[end_of_string], ((char **)args->arg_Arguments[argument_number].ta_CommandLineArg)[end_of_string]);
        						}
        						arg_values[argument_number].Ptr = args->arg_Arguments[argument_number].ta_CommandLineArg;
							}
							else
							{
								arg_values[argument_number].Ptr = args->arg_Arguments[argument_number].ta_CommandLineArg;
								d1(printf(__FILE__ "/%s/%d: String=<%s>\n", __func__, __LINE__, (char *)arg_values[argument_number].Ptr));
							}
							break;
							
						case TATF_NUMBER:
							d1(printf(__FILE__ "/%s/%d: Number\n", __func__, __LINE__));
							if(NULL != arg_values[argument_number].Ptr)
							{
								d1(printf(__FILE__ "/%s/%d: Pointer to long is valid\n", __func__, __LINE__));
								
								*((long *)arg_values[argument_number].Ptr) = strtol(args->arg_Arguments[argument_number].ta_CommandLineArg, &arg_end, 10);
								d1(printf(__FILE__ "/%s/%d: Converted as %ld\n", __func__, __LINE__, *((long *)arg_values[argument_number].Ptr)));
								
								if(*arg_end != '\0')
								{
									args->arg_Error = ARG_ERROR_INVALID_NUMBER;
	   								args->arg_ErrorPosition = argument_number;
								}
							}
							break;
							
						case TATF_BOOLEAN:
							arg_values[argument_number].Long = 1;
							d1(printf(__FILE__ "/%s/%d: Boolean\n", __func__, __LINE__));
							break;
							
						case TATF_TOGGLE:
							arg_values[argument_number].Long = !(arg_values[argument_number].Long);
							d1(printf(__FILE__ "/%s/%d: Toggle, new status = %s\n", __func__, __LINE__, arg_values[argument_number].Long ? "true" : "false"));
							break;
							
						default:
							d1(printf(__FILE__ "/%s/%d: Unknown type, freaking out\n", __func__, __LINE__));
							args->arg_Error = ARG_ERROR_TYPE_COMBINATION;
	   						args->arg_ErrorPosition = argument_number;
							break;
					}
				}
   			}

		}
		else
		{
			/* Failed to allocate memory for the copy of the template string */
			args->arg_Error = ARG_ERROR_ALLOCATION;
			args->arg_ErrorPosition = -1;
		}
		
		/* Get the real return code from the args structure */
		error = args->arg_Error;
	}

	d1(printf(__FILE__ "/%s/%d: Exit with error=%ld\n", __func__, __LINE__, error));

	return(error);
}


/*
 *	Function:		ra_ArgumentType
 *	Synopsis:		type = ra_ArgumentType(args, arg_number)
 *	Parameters:		struct Arguments *args	- Address of a valid base data structure
 *					long arg_number			- The argument number in the template to figure out the type for
 *	Returns:		unsigned long type - The type bit mask for the argument in the template
 *	Description:	Figures out the type of an argument string in a template from the
 *					modifiers in the string. Removes the modifiers to leave just the
 *					argument names and checks for valid combinations of argument types.
 */
static unsigned long	ra_ArgumentType(struct Arguments *args, long arg_number)
{
	unsigned long	type = 0;				/* The type we have figured out the argument to be (defaults to string, for no modifiers) */
	char			*type_start;			/* Pointer to the first character in the type specifier */
	char			*type_end;				/* Pointer to the end of each type specifier */
	int				type_number;			/* Used as a loop counter to check what type was found */

	if(args)
	{
		if(arg_number >= 0 && arg_number < args->arg_NumberArgs && args->arg_Arguments[arg_number].ta_ArgumentName)
		{
			/* Find first type modifier escape character and replace with null to remove them from names */
			type_end = ra_mystrtok(args->arg_Arguments[arg_number].ta_ArgumentName, TEMPLATE_TYPE, '\0');
			while(type_end)
			{
				/* Take copy of start of type and find and null terminate the end of it */
				type_start = type_end + 1;
				type_end = ra_mystrtok(NULL, TEMPLATE_TYPE, '\0');
				
				/* Search the allowable types array and set the bit masks if the types match */
				for(type_number = TATB_NUMBER; type_number <= TATB_EOL; type_number++)
				{
					if(strcmpi(type_start, ta_types[type_number]) == 0)
					{
						type |= (1 << type_number);
					}
				}
			}

			/* Make sure we have a basic type set */
			if((type & (TATF_STRING|TATF_NUMBER|TATF_BOOLEAN|TATF_TOGGLE)) == 0)
			{
				type |= TATF_STRING;
			}
			
			/* Check for invalid flag combinations. There should only be one of the four
			 * basic types specified, and you the EOL type can only be validly combined
    		 * with the string, keyword and required types.
       		 */
			if(ra_nbits(type & (TATF_STRING|TATF_NUMBER|TATF_BOOLEAN|TATF_TOGGLE)) != 1 || ((type & TATF_EOL) && ra_nbits(type & (TATF_NUMBER|TATF_BOOLEAN|TATF_TOGGLE|TATF_MULTIPLE)) != 0) || (ra_nbits(type & (TATF_NUMBER|TATF_BOOLEAN|TATF_TOGGLE|TATF_MULTIPLE)) > 1))
			{
				args->arg_Error = ARG_ERROR_TYPE_COMBINATION;
				args->arg_ErrorPosition = arg_number;
			}
		}
		else
		{
			args->arg_Error = ARG_ERROR_INVALID_ARGS;
			args->arg_ErrorPosition = -1;
		}
	}
	
	return(type);
}


/*
 *	Function:		ra_SeparateAlternatives
 *	Synopsis:		ra_SeparateAlternatives(args, arg_number)
 *	Parameters:		struct Arguments *args	- Address of a valid base data structure
 *											  which is ready to have the alternative argument names extracted.
 *					long arg_number			- The argument which is currently being processed
 *	Returns:		Nothing, although the error field may be filled in and the template alternative
 *					names should be.
 *	Description:	Takes an Argument base data structure which is ready for the
 *					argument alternative names to be separated (arguments separated,
 *					types extracted, alternatives counted and memory allocated)
 *					and splits the argument name string into the alternatives array.
 */
static void	ra_SeparateAlternatives(struct Arguments *args, long arg_number)
{
	char	*end_name;		/* Pointer to the equals symbols used to separate the alternative argument names */
	int		alt_number = 0;	/* Counter for the alternative names */
	
	/* Make sure all the parameters to this function are OK */
	if(args && ARG_ERROR_NONE == args->arg_Error)
	{
		/* And the stuff in the base data structures */
		if(arg_number >= 0 && arg_number < args->arg_NumberArgs && args->arg_Arguments && args->arg_Arguments[arg_number].ta_ArgumentName && args->arg_Arguments[arg_number].ta_Alternatives)
		{
			/* Find end of main argument name and null terminate it if not already done so */
			end_name = ra_mystrtok(args->arg_Arguments[arg_number].ta_ArgumentName, TEMPLATE_ALTERNATIVE, '\0');
			
			/* While we have found alternative form separators... */
			while(end_name)
			{
				/* Take a copy of the pointer to the start of the string, and null terminate it */
				args->arg_Arguments[arg_number].ta_Alternatives[alt_number] = end_name + 1;
				end_name = ra_mystrtok(NULL, TEMPLATE_ALTERNATIVE, '\0');
				
				/* Remove any whitespace from the start of the string and increase the alternative count if it is not empty */
				args->arg_Arguments[arg_number].ta_Alternatives[alt_number] += strspn(args->arg_Arguments[arg_number].ta_Alternatives[alt_number], WHITESPACE);
				if(strlen(args->arg_Arguments[arg_number].ta_Alternatives[alt_number]) > 0)
				{
					alt_number++;
				}
			}
		}
		else
		{
			/* Set the error value of the base data structure to show there was
			 * a problem in this function.
			 */
			args->arg_Error = ARG_ERROR_INVALID_ARGS;
			args->arg_ErrorPosition = arg_number;
		}
	}
}


/*
 *	Function:		ra_FindArgumentName
 *	Synopsis:		found_argument = ra_FindArgumentName(args, start_from, name)
 *	Parameters:		struct Arguments *args				- Address of the main Arguments structure in which the parsed template argument is searched for
 *					struct TemplateArgument *start_from	- The address of the parsed template argument from which to start searching from, or NULL to search from the start
 *					char *name							- The name string of the argument to find
 *	Returns:		struct TemplateArgument *found_argument - The address of the parsed template argument information or NULL if it could not be found
 *	Description:	Searches for an argument name in the parsed template strings.
 */
static struct TemplateArgument *ra_FindArgumentName(struct Arguments *args, struct TemplateArgument *start_from, char *name)
{
	struct TemplateArgument	*found_argument = NULL;
	long					current_argument;
	long					current_alternative;

	/* Make sure the pointers into the function are OK */	
	if(args)
	{
		/* And the necessary bits inside the base structure */
		if(args->arg_Arguments && name)
		{
			/* Skip any symbols at the start of the name */
			name += strspn(name, "-/");
			
			/* Loop through all the arguments, checking for an argument whose type matches the given type (mask/pattern)
			 * If a starting TemplateArgument has been specified, start the search from there
			 */
			if(start_from)
			{
				current_argument = ((unsigned long)start_from - (unsigned long)args->arg_Arguments) / sizeof(struct TemplateArgument);
			}
			else
			{
				current_argument = 0;
			}
			
			for(; current_argument < args->arg_NumberArgs && NULL == found_argument; current_argument++)
			{
				/* Check the (case insensitive) main name of the argument */
				if(stricmp(args->arg_Arguments[current_argument].ta_ArgumentName, name) == 0)
				{
					found_argument = &args->arg_Arguments[current_argument];
				}
				else
				{
					/* Check any alternatives */
					for(current_alternative = 0; current_alternative < args->arg_Arguments[current_argument].ta_NumberAlternatives && NULL == found_argument; current_alternative++)
					{
						if(stricmp(args->arg_Arguments[current_argument].ta_Alternatives[current_alternative], name) == 0)
						{
							found_argument = &args->arg_Arguments[current_argument];
						}
					}
				}
			}
		}
		else
		{
			/* Base data structure is incomplete, so return something to tell the calling function */
			args->arg_Error = ARG_ERROR_INVALID_ARGS;
			args->arg_ErrorPosition = -1;
		}
	}
	
	/* And return the argument which was found, if any */
	return(found_argument);
}


/*
 *	Function:		ra_FindArgumentType
 *	Synopsis:		found_argument = ra_FindArgumentType(args, start_from, mask, pattern)
 *	Parameters:		struct Arguments *args				- The address of the main Arguments structure in which to search for the parsed template argument
 *					struct TemplateArgument *start_from	- Address of the template argument to start searching from, or NULL to search from the start
 *					unsigned long mask					- Bit-mask to be applied to the template type flags when performing comparisons
 *					unsigned long pattern				- The bit mask to compare to the type flags in the template argument (after masking), for a successful match
 *	Returns:		struct TemplateArgument *found_argument - The address of the parsed template argument information or NULL if it could not be found
 *	Description:	Searches for an argument with the specified flags (masked out
 *					with the flags and compared to the pattern).
 */
static struct TemplateArgument *ra_FindArgumentType(struct Arguments *args, struct TemplateArgument *start_from, unsigned long mask, unsigned long pattern)
{
	struct TemplateArgument	*found_argument = NULL;
	long					current_argument;

	/* Make sure the pointers into the function are OK */	
	if(args)
	{
		/* And the necessary bits inside the base structure */
		if(args->arg_Arguments)
		{
			/* Loop through all the arguments, checking for an argument whose type matches the given type (mask/pattern)
			 * If a starting TemplateArgument has been specified, start the search from there
			 */
			if(start_from)
			{
				current_argument = ((unsigned long)start_from - (unsigned long)args->arg_Arguments) / sizeof(struct TemplateArgument);
			}
			else
			{
				current_argument = 0;
			}
			
			for(; current_argument < args->arg_NumberArgs && NULL == found_argument; current_argument++)
			{
				if((args->arg_Arguments[current_argument].ta_Type & mask) == pattern)
				{
					found_argument = &args->arg_Arguments[current_argument];
				}
			}
		}
		else
		{
			/* Base data structure is incomplete, so return something to tell the calling function */
			args->arg_Error = ARG_ERROR_INVALID_ARGS;
			args->arg_ErrorPosition = -1;
		}
	}
	
	/* And return the argument which was found, if any */
	return(found_argument);
}


/*
 *	Function:		ra_ScanForPrompt
 *	Synopsis:		error = ra_ScanForPrompt(args, template_string, command_line)
 *	Parameters:		struct Arguments *args	- The address of the main Arguments structure in which to search for the parsed template argument
 *					char *template_string	- Address of the start of a character string containing the command line template
 *					char *command_line		- Address of the start of a character string containing the program's command line
 *	Returns:		long error - Flag showing whether there were any errors during the parsing of the command line
 *	Description:	Scans the command line for a question mark, which is used to
 *					signify that the user wants to be prompted by the program.
 *
 *					The user is prompted with the template string which is written
 *					to stdout.
 */
static long	ra_ScanForPrompt(struct Arguments *args, char *template_string, char *command_line)
{
	long	error = ARG_ERROR_NONE;
	char	*arg_start;
	char	*arg_end;
	int		end_of_string = 0;
	
	if(NULL != args)
	{
		if(NULL != command_line)
		{
			arg_end = command_line;
			do
			{
				/* Find start of argument - any non whitespace */
				arg_start = arg_end + strspn(arg_end, WHITESPACE);
				if(!*arg_start)
				{
					end_of_string = 1;
					arg_start = arg_end;
				}
				
				/* Find end of argument - any whitespace, quote or equals sign */
				if('\"' == *arg_start)
				{
					/* If argument starts with a quote, we need to search for the terminating one */
					arg_end = (arg_start + 1) + strcspn(arg_start + 1, "\"");
				}
				else
				{
					/* Otherwise we look for whitespace or equals */
					arg_end = (arg_start + 1) + strcspn(arg_start + 1, WHITESPACE "=");
				}
				
				/* Check for the end of the string */
				if(!*arg_end)
				{
					end_of_string = 1;
				}

				d1({
					char	fn[256];
					char	sp[256];
					
					sprintf(sp, "%%.%ds", arg_end - arg_start);
					sprintf(fn, sp, arg_start);
					printf(__FILE__ "/%s/%d: Found argument <%s> with length %d\n", __func__, __LINE__, fn, arg_end - arg_start);
				})

				if(strncmp("?", arg_start, (arg_end - arg_start)) == 0)
				{	
					char	prompted_input[PROMPT_STRING_LENGTH] = {0};
					char	*new_string;
					
					end_of_string = 1;
					
					printf("%s: ", template_string);
					if(getline(prompted_input, PROMPT_STRING_LENGTH-1) > 0)
					{
						new_string = calloc(strlen(args->arg_CommandLine) + strlen(prompted_input) + 2, sizeof(char));
						if(NULL != new_string)
						{
							/* Join current argument string onto newly entered part, removing the ? character we found */
							strncpy(new_string, args->arg_CommandLine, arg_start - args->arg_CommandLine);
							sprintf(&new_string[strlen(new_string)], "%s %s", arg_end, prompted_input);
							free(args->arg_CommandLine);
							args->arg_CommandLine = new_string;
						}
						else
						{
							error = args->arg_Error = ARG_ERROR_ALLOCATION;
							args->arg_ErrorPosition = -1;			
						}
					}
					else
					{
						/* No new string entered, but we don't want the question mark */
						*arg_start = ' ';
					}
				}
			} while(!end_of_string);
		}
		else
		{
			error = args->arg_Error = ARG_ERROR_INVALID_ARGS;
			args->arg_ErrorPosition = -1;			
		}
	}
	else
	{
		error = ARG_ERROR_INVALID_ARGS;
	}
	
	return(error);
}


/*
 *	Function:		ra_SkipCommandName
 *	Synopsis:		command_name_length = ra_SkipCommandName(command_string)
 *	Parameters:		char *command_string - Pointer to the start of the string found on the command line
 *	Returns:		The length of the string to skip which represents the executable name and any trailling whitespace
 *	Description:	Tries to find the name of the executable as called in the command
 *					line so that it can be skipped and the only parsing performed after
 *					that is on the command line parameters. This is done by getting
 *					the module name and matching it to the start of the command
 *					line. If this fails, the first string on the command line is
 *					taken as the command name - either up to the first whitespace
 *					or a terminating quote if the string starts with a quote.
 */
static long		ra_SkipCommandName(char *command_string)
{
	long	command_name_length = 0;
	char	exe_name[EXE_NAME_LENGTH];
	
	d1(printf(__FILE__ "/%s/%d: Entered command_string=<%s>\n", __func__, __LINE__, command_string));
	
	if(command_string)
	{
		/* Try to skip past the executable filename which will be at the start of the string */
		if(GetModuleFileName(NULL, exe_name, EXE_NAME_LENGTH))
		{
			d1(printf(__FILE__ "/%s/%d: Executable name=<%s>\n", __func__, __LINE__, exe_name));
					
			if(_strnicmp(exe_name, command_string, strlen(exe_name)) == 0)
			{
				// Move pointer past exe filename and any whitespace
				command_name_length = strlen(exe_name);
			}
			else
			{
				// Try to parse end of executable name manually
				if('\"' == *command_string)
				{
					// Find terminating quote
					command_name_length = 1 + strcspn(&command_string[1], "\"");
					d1(printf(__FILE__ "/%s/%d: Length of quote enclosed executable name=%ld\n", __func__, __LINE__, command_name_length));
				}
				else
				{
					// Find next whitespace
					command_name_length = strcspn(command_string, WHITESPACE);
					d1(printf(__FILE__ "/%s/%d: Length of unquoted executable name=%ld\n", __func__, __LINE__, command_name_length));
				}
			}
		}
		else
		{
			// Try to parse end of executable name manually
			if('\"' == *command_string)
			{
				// Find terminating quote
				command_name_length = 1 + strcspn(&command_string[1], "\"");
				d1(printf(__FILE__ "/%s/%d: Length of quote enclosed executable name=%ld\n", __func__, __LINE__, command_name_length));
			}
			else
			{
				// Find next whitespace
				command_name_length = strcspn(command_string, WHITESPACE);
				d1(printf(__FILE__ "/%s/%d: Length of unquoted executable name=%ld\n", __func__, __LINE__, command_name_length));
			}
		}

		// Skip remaining whitespace up to start of first argument
		command_name_length += strspn(&command_string[command_name_length], WHITESPACE);

		// Something to show what we've removed
		d1({
			int	i;
			
			printf(__FILE__ "/%s/%d: String removed from command line=<", __func__, __LINE__);
   			for(i = 0; i < command_name_length; i++)
      		{
        		printf("%c", command_string[i]);
			}
   			printf(">\n");
		});

	}
	
	return(command_name_length);
}


/*
 *	Function:		ra_nbits
 *	Synopsis:		n = ra_nbits(x)
 *	Parameters:		unsigned long x - The bit pattern you want to count the number of set bits in
 *	Returns:		long n - The number of bits in the bit pattern which were set to 1
 *	Description:	Counts the number of bits which are set to 1 in a bit pattern.
 *	Example:
 *	{
 *		long	ra_nbits(unsigned long x);
 *		int i;
 *		
 *		for(i = 0 ; i < 20; i++)
 *		{
 *			printf("Bits in %d is %ld\n", i, ra_nbits(i));
 *		}
 *	}
 */
static inline long	ra_nbits(unsigned long x)
{
	long	n = 0;
	while(x)
	{
		n++;
		x &= (x-1);
	}
	
	return(n);
}


/*
 *	Function:		ra_log2
 *	Synopsis:		n = ra_log2(x)
 *	Parameters:		unsigned long x - The value of which you want the logarithm to the base 2
 *	Returns:		long n - log2(x) as an integer or -1 if zero was passed
 *	Description:	Calculates the logarithm to the base 2 of an integer using
 *					bit shifting. The value is calculated to the nearest integer
 *					*lower* than the exact value. So, for example, if you passed
 *					127 you would get a result of 6, even though 128 (which is the
 *					nearest integer to the original value) would result in 7.
 */
static inline long	ra_log2(unsigned long x)
{
	long	n = -1;
	
	while(x)
	{
		d1(printf(__FILE__ "/%s/%d: x=%lu n=%ld\n", __func__, __LINE__, x, n));
		x = x >> 1;
		n++;
	}
	
	return(n);
}

