/*
 *	Filename:		readargs.h
 *	Description:	Definitions for command line parsing functions, structures and constants
 */

#ifndef READARGS_H
#define READARGS_H

/* Useful when an item in an array has to be accessable as a pointer or an integer
 * and the possibility exists for pointers and integers to be different sizes.
 * Shouldn't happen in Win32, but perhaps someday...
 */
union LongPtrUnion
{
	unsigned long	Long;
	void			*Ptr;
};
typedef union LongPtrUnion LongPtrType;


/* Private - the internal structure of a single argument in a template */
struct TemplateArgument
{
	char			*ta_ArgumentName;		/* Pointer to a string containing the argument main name */
	unsigned long	ta_Type;				/* Type of argument, built up from the flags specified below */ 
	long			ta_NumberAlternatives;	/* Number of alternative strings for this argument */
	char			**ta_Alternatives;		/* Array of strings for each alternative form */
	char			*ta_CommandLineArg;		/* Pointer to the section of the command line which matches the template */
};


/* Types of arguments in templates */
enum
{
	TATB_STRING = 0,	/* Argument is a string, no type modifier for the template, set as a pointer to the string or NULL for not specified in the user array */
	TATB_NUMBER,		/* Argument is an integer number (up to long size), /N template type modifier, set as a pointer to the long variable to receive the number or NULL for not specified in the user array */
	TATB_BOOLEAN,		/* Argument is a boolean, /B template type modifier, boolean state of entry in user array indicates whether this argument was specified or not */
	TATB_TOGGLE,		/* Argument is a toggle (boolean), /T template type modifier, boolean state of entry in user array is toggled if argument is specified */
	TATB_KEYWORD,		/* Argument will only be recognised if preceeded by argument name on command line, /K template type modifier */
	TATB_REQUIRED,		/* Argument is required for command line parse to be successful, /R template type modifier */
	TATB_MULTIPLE,		/* Allows the user application to get access to whitespace separated arguments as separate array items (user array entry will be pointer to NULL terminated values) which have not been picked up as specific arguments, /M template type modifier */
	TATB_EOL			/* Allows the user application to get access to the remainder of the command line (all recognised parameters removed) as a single string, /E template type modifier, no other arguments or types specified after this in the template will be recognised */
};


/* And the corresponding bit-masks */
#define TATF_STRING		(1 << TATB_STRING)
#define TATF_NUMBER		(1 << TATB_NUMBER)
#define TATF_BOOLEAN	(1 << TATB_BOOLEAN)
#define TATF_TOGGLE		(1 << TATB_TOGGLE)
#define TATF_KEYWORD	(1 << TATB_KEYWORD)
#define TATF_REQUIRED	(1 << TATB_REQUIRED)
#define TATF_MULTIPLE	(1 << TATB_MULTIPLE)
#define TATF_EOL		(1 << TATB_EOL)


/* And the characters which are used to represent them */
#define TATS_STRING		""
#define TATS_NUMBER		"N"
#define TATS_BOOLEAN	"B"
#define TATS_TOGGLE		"T"
#define TATS_KEYWORD	"K"
#define TATS_REQUIRED	"R"
#define TATS_MULTIPLE	"M"
#define TATS_EOL		"E"
extern const char * const ta_types[];


/* A couple of other type values which are used during the extraction of argument names
 * and values from the command line. Note that these are values, to be used as
 * numbers by themselves!
 */
#define TATB_NONE		-1
#define TATB_VALUE		-2


/* Private contents, but to be used in applications as the base structure for
 * handling command line arguments
 */
struct Arguments
{
	short					arg_Error;			/* Shows the state of any error during parsing of template or command line - see below */
	short					arg_ErrorPosition;	/* The ordinal argument number in which the error occurred, for either template or command line (note that command line position does not necessarily correspond 1:1 with template position) */
	char					*arg_Template;		/* A copy of the template, butchered into our own internal format */
	char					*arg_CommandLine;	/* A copy of the command line, butchered into an internal format */
	long					arg_NumberArgs;		/* Number of arguments to be recognised in template */
	struct TemplateArgument	*arg_Arguments;		/* Array for the number of arguments in the template */
};


/* Error values */
#define ARG_ERROR_INVALID_NUMBER			-10		/* Tried to convert a string into a number (for a numeric argument type) but failed */
#define ARG_ERROR_REQUIRED_MISSING			-9		/* Required argument missing from command line, error position tells you at which one (the template position in this case) */
#define ARG_ERROR_MULTIPLES					-8		/* Signifies that there are multiple "multiple" argument types in the template */
#define ARG_ERROR_TYPE_COMBINATION			-7		/* Invalid combination of argument types in template, error position tells you at which one */
#define ARG_ERROR_UNRECOGNISED_TYPE			-6		/* Unrecognised type of argument in template, error position tells you at which one */
#define ARG_ERROR_ALTERNATIVE_SEPARATION	-5		/* Failed during the separation of argument alternative forms, error position tells you at which one */
#define ARG_ERROR_ARGUMENT_SEPARATION		-4		/* Failed during the separation of arguments, error position tells you at which one */
#define ARG_ERROR_TEMPLATE_ALLOCATION		-3		/* Failed to allocate memory during template parsing - different from failing to allocate the base block */
#define ARG_ERROR_ALLOCATION				-2		/* Failed to allocate memory when taking working copies of data */
#define ARG_ERROR_INVALID_ARGS				-1		/* Invalid arguments to (internal) functions */
#define ARG_ERROR_NONE						0		/* No errors */


/* Public functions to be used to parse a command line and then free any allocations which
 * were made during the parsing.
 */
struct Arguments	*ReadArguments(char *template_string, LongPtrType *arg_values, char *alternative_arguments);
void				FreeArguments(struct Arguments *args);
char				*FindArgumentByNumber(char *template_string, int argument_number, char *buffer, unsigned long buffer_length);


#endif /* READARGS_H */

