/*
 *	Filename:		clutils.c
 *	Description:	Useful command line routines
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

#include "clutils.h"


/*
 *	Function:		clrscr
 *	Synopsis:		clrscr()
 *	Parameters:		None
 *	Returns:		Nothing
 *	Description:	Clears a win32 console display. Should also be capable of handling
 *					consoles which have been scrolled or resized.
 */
void	clrscr(void)
{
	int							position;					/* Loop counter for clearing rows and columns */
	HANDLE						console;					/* Handle to the OS created console window */
	CONSOLE_SCREEN_BUFFER_INFO	console_screen;				/* Screen buffer for the console display */
	COORD						cursor_position;			/* X/Y co-ordinates used while clearing the console */
	unsigned long				written;					/* A place to get the result of how many characters were written to the console */
	char						*blank_row;					/* Pointer to a string of spaces (dynamically allocated) to clear the display */
	int							total_rows, total_columns;	/* The area of the console to be cleared */
    DWORD						old_mode;					/* Stores the old mode of operation of the console */
    int							old_mode_ok;				/* Shows if the old mode could be found successfully */
	COORD						old_position;				/* X/Y co-ordinates of the cursor position before clearing the screen */

	/* Get access to the current console, check that it is OK */
	console = GetStdHandle(STD_OUTPUT_HANDLE);
	if(INVALID_HANDLE_VALUE != console)
	{
		/* Try to get the current mode of the console. If successful, turn off end of line wrapping */
	    old_mode_ok = GetConsoleMode(console, &old_mode);
    	if(old_mode_ok)
	    {
    		SetConsoleMode(console, old_mode & ~ENABLE_WRAP_AT_EOL_OUTPUT);
	    }

		/* Try to get the information about the screen buffer for the console */
		if(GetConsoleScreenBufferInfo(console, &console_screen))
		{
			/* Set the default console colours */
			ConsoleColour(CONSOLE_COLOR_WHITE, CONSOLE_COLOR_BLACK);
			
			/* Record the current cursor position and area to be cleared */
		    old_position.X = 0;//console_screen.srWindow.Left;//console_screen.dwCursorPosition.X;
    		old_position.Y = 0;//console_screen.srWindow.Top;//console_screen.dwCursorPosition.Y;
			total_rows = console_screen.srWindow.Bottom - old_position.Y + 1;
			total_columns = console_screen.srWindow.Right - old_position.X + 1;

			/* Allocate a full row to be able to blank entire lines at once */
			blank_row = (char *)malloc(total_columns + 1);
			if(blank_row)
			{
				/* If row was allocated, fill it with spaces snd terminate with null */
				for(position = 0; position < total_columns; position++)
				{
					blank_row[position] = ' ';
				}
				blank_row[total_columns] = '\0';
				
				/* Take a working copy of the cursor X position */
        		cursor_position.X = old_position.X;

				/*	Loop for each row, calculating the new Y position, moving there and
    			 *	writing the line of blanks
    			 */
				for(position = 0; position < total_rows; position++)
				{
					cursor_position.Y = position + old_position.Y;
					SetConsoleCursorPosition(console, cursor_position);
					WriteConsole(console, blank_row, total_columns, &written, 0);
				}

				/* Free the memory allocated for the line of spaces */
				free(blank_row);
			}
			else
			{
				const char	* const	blank_space = " ";		/* Used to clear single spaces within the following loops */
				
				/* Clear the screen the hard way using single characters and for loops for each column/row */
				for(cursor_position.Y = old_position.Y; cursor_position.Y < (old_position.Y + total_rows); cursor_position.Y++)
				{
					for(cursor_position.X = old_position.X; cursor_position.X < (old_position.X + total_rows); cursor_position.X++)
					{
						SetConsoleCursorPosition(console, cursor_position);
						WriteConsole(console, blank_space, 1, &written, 0);
					}
				}
			}
	
			/* Restore cursor to the position it was at before we cleared the screen */
			old_position.X = 0;
			old_position.Y = 0;
			SetConsoleCursorPosition(console, old_position);
		}

		/* If we were able to find out the console mode at the start of this function, restore it */
	    if(old_mode_ok)
    	{
    		SetConsoleMode(console, old_mode);
	    }
 	}
}


/*
 *	Function:		ConsoleLocate
 *	Synopsis:		ConsoleLocate(x, y)
 *	Parameters:		int x - Column position to move the cursor to
 *					int y - Row position to move the cursor to
 *	Returns:		Nothing
 *	Description:	Moves the current position of the text mode cursor in a Win32
 *					console to those specified in the parameters, with origin of
 *					0,0 being the top-left of the display.
 */
void	ConsoleLocate(int x, int y)
{
	COORD	cursor_position;	/* The OS structure used for setting console cursor positions */
	HANDLE	console;			/* Handle to the current standard output (stdout - hopefully a console) */
	
	/* Copy the user specified X/Y co-ordinates to the required structure */
	cursor_position.X = x;
	cursor_position.Y = y;
	
	/* Try to get access to the current output handle - if successful, set the position */
	console = GetStdHandle(STD_OUTPUT_HANDLE);
	if(INVALID_HANDLE_VALUE != console)
	{
		SetConsoleCursorPosition(console, cursor_position);
	}
}


/*
 *	Function:		ConsolePosition
 *	Synopsis:		ConsolePosition(&x, &y)
 *	Parameters:		int *x - The address of an integer in which to store the X co-ordinate (or NULL if you do not want it stored)
 *					int *y - The address of an integer in which to store the Y co-ordinate (or NULL if you do not want it stored)
 *	Returns:		Nothing (although values are returned through the pointers in the parameters)
 *	Description:	Finds out the current cursor position in a Win32 console.
 *					The origin 0,0 starts at the top-left of the window.
 */
void	ConsolePosition(int *x, int *y)
{
	HANDLE	console;							/* Handle to the current standard output (stdout - hopefully a console) */
	CONSOLE_SCREEN_BUFFER_INFO	console_screen;	/* Screen buffer for the console display */
	
	/* Try to get access to the current output handle - if successful, set the position */
	console = GetStdHandle(STD_OUTPUT_HANDLE);
	if(INVALID_HANDLE_VALUE != console)
	{
		/* Try to get the information about the screen buffer for the console */
		if(GetConsoleScreenBufferInfo(console, &console_screen))
		{
			/* If the user has given us a valid pointer to the X storage space, retrieve it from the screen buffer info */
			if(x)
			{
			    *x = console_screen.dwCursorPosition.X;
		    }
		    
			/* If the user has given us a valid pointer to the Y storage space, retrieve it from the screen buffer info */
		    if(y)
		    {
	    		*y = console_screen.dwCursorPosition.Y;
    		}
		}
	}
}


static int	fg_colours[] = {0	/* black */,
 							FOREGROUND_BLUE	/* Blue */,
 							FOREGROUND_RED /* Red */,
 							FOREGROUND_GREEN /* Green */,
 							FOREGROUND_RED | FOREGROUND_BLUE	/* magenta */,
 							FOREGROUND_GREEN | FOREGROUND_BLUE	/* cyan */,
 							FOREGROUND_RED | FOREGROUND_GREEN	/* yellow */,
 							FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE	/* light grey */,
 							FOREGROUND_INTENSITY	/* dark grey */,
 							FOREGROUND_BLUE | FOREGROUND_INTENSITY	/* Blue */,
 							FOREGROUND_RED | FOREGROUND_INTENSITY /* Red */,
 							FOREGROUND_GREEN | FOREGROUND_INTENSITY /* Green */,
 							FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_INTENSITY	/* magenta */,
 							FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY	/* cyan */,
 							FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_INTENSITY	/* yellow */,
 							FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY	/* Bright white */};

static int	bg_colours[] = {0	/* black */,
 							BACKGROUND_BLUE	/* Blue */,
 							BACKGROUND_RED /* Red */,
 							BACKGROUND_GREEN /* Green */,
 							BACKGROUND_RED | BACKGROUND_BLUE	/* magenta */,
 							BACKGROUND_GREEN | BACKGROUND_BLUE	/* cyan */,
 							BACKGROUND_RED | BACKGROUND_GREEN	/* yellow */,
 							BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE	/* light grey */,
 							BACKGROUND_INTENSITY	/* dark grey */,
 							BACKGROUND_BLUE | BACKGROUND_INTENSITY	/* Blue */,
 							BACKGROUND_RED | BACKGROUND_INTENSITY /* Red */,
 							BACKGROUND_GREEN | BACKGROUND_INTENSITY /* Green */,
 							BACKGROUND_RED | BACKGROUND_BLUE | BACKGROUND_INTENSITY	/* magenta */,
 							BACKGROUND_GREEN | BACKGROUND_BLUE | BACKGROUND_INTENSITY	/* cyan */,
 							BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_INTENSITY	/* yellow */,
 							BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE | BACKGROUND_INTENSITY	/* Bright white */};


/*
 *	Function:		ConsoleColour		
 *	Synopsis:		ConsoleColour(foreground, background)
 *	Parameters:		ConsoleColourType foreground - the colour to draw the foreground of the characters in
 *					ConsoleColourType background - the colour to draw the background of the characters in
 *	Returns:		Nothing
 *	Description:	Sets the foreground and background colour in which to draw all 
 *					following characters. The foreground is the colour in which the
 *					character will be drawn. The background is all the surrounding
 *					area. The colour comes from the enumerated values defined by
 *					ConsoleColourType.
 */
void	ConsoleColour(ConsoleColourType foreground, ConsoleColourType background)
{
	HANDLE						console;		/* Handle to the console to set the colours for */
	CONSOLE_SCREEN_BUFFER_INFO	console_screen;	/* Screen buffer for the console display (to get the unchanged colours) */
	int							fg, bg;			/* The actual colour values that will be set to the console */

	/* Don't need to bother doing anything if both colours are not to be changed */
	if(CONSOLE_COLOR_NOCHANGE != foreground || CONSOLE_COLOR_NOCHANGE != background)
	{
		/* Try to get access to the current output handle - if successful, set the character colour */
		console = GetStdHandle(STD_OUTPUT_HANDLE);
		if(INVALID_HANDLE_VALUE != console)
		{
			/* Set the appropriate foreground bitmask */
			if(CONSOLE_COLOR_NOCHANGE == foreground)
			{
				/* Try to get the foreground colour from the screen info (mask out all but
				 * the foreground bits) or set to grey as a fallback.
				 */
				if(GetConsoleScreenBufferInfo(console, &console_screen))
				{
					fg = console_screen.wAttributes & (FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
				}
				else
				{
					fg = fg_colours[CONSOLE_COLOR_GREY];
				}
			}
			else
			{
				fg = fg_colours[foreground];
			}
		
			/* Set the appropriate background bitmask */
			if(CONSOLE_COLOR_NOCHANGE == background)
			{
				/* Try to get the background colour from the screen info (mask out all but
				 * the background bits) or set to black as a fallback.
				 */
				if(GetConsoleScreenBufferInfo(console, &console_screen))
				{
					bg = console_screen.wAttributes & (BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE | BACKGROUND_INTENSITY);
				}
				else
				{
					bg = bg_colours[CONSOLE_COLOR_BLACK];
				}
			}
			else
			{
				bg = bg_colours[background];
			}

			/* Set the colour to the console */
			SetConsoleTextAttribute(console, fg | bg);
		}
	}
}


/*
 *	Function:		ConsoleForeground
 *	Synopsis:		colour = ConsoleForeground()
 *	Parameters:		None
 *	Returns:		ConsoleColourType colour - The current foreground colour (index) of the console
 *	Description:	Finds the current foreground colour in the current standard output
 *					(which would need to be a Win32 console) as one of the enumerated
 *					names.
 */
ConsoleColourType	ConsoleForeground(void)
{
	HANDLE				console;						/* Handle to the console to get the foreground colour from */
	int					fg;								/* The actual colour value that will be got from the console */
	ConsoleColourType	colour = CONSOLE_COLOR_WHITE;	/* The enumerated value representing that colour */
	int					i;								/* Loop counter to cycle through the enumerated colours in the index */
	CONSOLE_SCREEN_BUFFER_INFO	console_screen;	/* Screen buffer for the console display (to get the unchanged colours) */

	/* Try to get access to the current output handle - if successful, set the character colour */
	console = GetStdHandle(STD_OUTPUT_HANDLE);
	if(INVALID_HANDLE_VALUE != console)
	{
		if(GetConsoleScreenBufferInfo(console, &console_screen))
		{
			/* Get the current colour from the console screen buffer */
			fg = console_screen.wAttributes & (FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
			
			/* Loop through all the defined colours until we find the one we are looking for, or reach the end of the array */
			for(i = CONSOLE_COLOR_BLACK; i <= CONSOLE_COLOR_BRIGHT_WHITE && fg != fg_colours[i]; i++)
			{
				;
			}
			
			/* Set the colour returned by this function to be the one we found with the above loop */
			if(i <= CONSOLE_COLOR_BRIGHT_WHITE)
   			{
      			colour = (ConsoleColourType) i;
         	} 
		}
	}
	
	return(colour);
}


/*
 *	Function:		ConsoleBackground
 *	Synopsis:		colour = ConsoleBackground()
 *	Parameters:		None
 *	Returns:		ConsoleColourType colour - The current background colour (index) of the console
 *	Description:	Finds the current background colour in the current standard output
 *					(which would need to be a Win32 console) as one of the enumerated
 *					names.
 */
ConsoleColourType	ConsoleBackground(void)
{
	HANDLE						console;						/* Handle to the console to get the background colour from */
	int							bg;								/* The actual colour value that will be got from the console */
	ConsoleColourType			colour = CONSOLE_COLOR_BLACK;	/* The enumerated value representing that colour */
	int							i;								/* Loop counter to cycle through the enumerated colours in the index */
	CONSOLE_SCREEN_BUFFER_INFO	console_screen;					/* Screen buffer for the console display (to get the unchanged colours) */

	/* Try to get access to the current output handle - if successful, set the character colour */
	console = GetStdHandle(STD_OUTPUT_HANDLE);
	if(INVALID_HANDLE_VALUE != console)
	{
		if(GetConsoleScreenBufferInfo(console, &console_screen))
		{
			/* Get the current colour from the console screen buffer */
			bg = console_screen.wAttributes & (BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE | BACKGROUND_INTENSITY);
			
			/* Loop through all the defined colours until we find the one we are looking for, or reach the end of the array */
			for(i = CONSOLE_COLOR_BLACK; i <= CONSOLE_COLOR_BRIGHT_WHITE && bg != bg_colours[i]; i++)
			{
				;
			}
			
			/* Set the colour returned by this function to be the one we found with the above loop */
			if(i <= CONSOLE_COLOR_BRIGHT_WHITE)
   			{
      			colour = (ConsoleColourType) i;
         	} 
		}
	}
	
	return(colour);
}


/*
 *	Function:		ConsoleInverse
 *	Synopsis:		ConsoleInverse(x, y, width, height)
 *	Parameters:		int x		- Horizontal position of the left edge of the block to invert
 *					int y		- Vertical position of the left edge of the block to invert
 *					int width	- Character width of the block to invert
 *					int height	- Row height of the block to invert
 *	Returns:		int success - Shows if the block was successfully inverted (true) or not (false).
 *								  A failure might also be a partial fill of the area.
 *	Description:	Highlights an area in a console window by inverting the video mode
 *					of the characters in that area.
 */
int					ConsoleInverse(int x, int y, int width, int height)
{
	HANDLE						console;						/* Handle to the console to get the background colour from */
	int							i;								/* Loop counter to cycle through the enumerated colours in the index */
	CONSOLE_SCREEN_BUFFER_INFO	console_screen;					/* Screen buffer for the console display (to get the unchanged colours) */
	int							success = 0;					/* Default return code is failure */
	COORD						highlight_position;				/* Co-ordinates to specify the current highlight position */
	DWORD						number_written;					/* Variable to recieve the number of attributes written to the console */
	WORD						*attributes;					/* Pointer to a block of memory to recieve and set the attributes */
	WORD						one_attribute;					/* And a fallback if we have to do it character by character */

	/* Try to get access to the current output handle - if successful, set the character colour */
	console = GetStdHandle(STD_OUTPUT_HANDLE);
	if(INVALID_HANDLE_VALUE != console)
	{
		if(GetConsoleScreenBufferInfo(console, &console_screen))
		{
			/* We got everything we need to access the console, so we should be looking at
			 * a successful return.
			 */
			success = 1;
			
			/* Don't need to do anything if the area is outside the console */
			if(width > 0 && height > 0 && x < console_screen.srWindow.Right && y < console_screen.srWindow.Bottom)
			{
				/* Correct the sizes so we do not go over the boundaries of the console,
				 * otherwise it might end up with stray highlights on the following rows
				 */
				if((x + width - 1) > console_screen.srWindow.Right)
				{
					width = console_screen.srWindow.Right - x + 1;
				}
				
				if((y + height - 1) > console_screen.srWindow.Bottom)
				{
					height = console_screen.srWindow.Bottom - y + 1;
				}

				attributes = (WORD *)malloc(sizeof(WORD) * width);
				if(NULL != attributes)
				{
					/* Do the area highlight for the area within the console */
					for(highlight_position.X = x, highlight_position.Y = y; highlight_position.Y < (y + height) && success; highlight_position.Y++)
					{
						success = ReadConsoleOutputAttribute(console, attributes, width, highlight_position, &number_written);
						if(number_written != width)
						{
							success = 0;
						}
						
						if(success)
						{
							for(i = 0; i < width; i++)
							{
								attributes[i] = attributes[i] ^ (FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY | BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE | BACKGROUND_INTENSITY);
							}
							
							success = WriteConsoleOutputAttribute(console, attributes, width, highlight_position, &number_written);
							if(number_written != width)
							{
								success = 0;
							}
						}
					} /* For each row in area */
					
					free(attributes);
				} /* attributes allocated OK */
				else
				{
					/* Need to highlight the area character by character */
					for(highlight_position.Y = y; highlight_position.Y < (y + height) && success; highlight_position.Y++)
					{
						for(highlight_position.X = x; highlight_position.X < (x + width) && success; highlight_position.X++)
						{
							success = ReadConsoleOutputAttribute(console, &one_attribute, 1, highlight_position, &number_written);
							if(number_written != 1)
							{
								success = 0;
							}
						
							if(success)
							{
								one_attribute = one_attribute ^ (FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY | BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE | BACKGROUND_INTENSITY);
								
								success = WriteConsoleOutputAttribute(console, &one_attribute, 1, highlight_position, &number_written);
								if(number_written != 1)
								{
									success = 0;
								}
							}
						} /* For each column in row */
					} /* For each row in area */
				} /* Failed to allocate attributes */
			} /* Area to be highlighted is within display */
		} /* Console screen buffer got */
	} /* Console handle successfully found */
	
	return(success);
}


/*
 *	Function:		is_modifier
 *	Synopsis:		modifier = is_modifier(kb)
 *	Parameters:		KEY_EVENT_RECORD *kb - Address of a structure which contains a keyboard event record
 *	Returns:		int modifier - Flag showing whether the keyboard event is a modifier key by itself
 *	Description:	This function is used to decide whether the keypress is one
 *					which the standard library getch() will recognise (false)
 *					or not (true).
 */
static int		is_modifier(KEY_EVENT_RECORD *kb)
{
	int		modifier = 0;
	
	if((29 == kb->wVirtualScanCode ||	/* Control */
 		42 == kb->wVirtualScanCode ||	/* Left shift */
   		54 == kb->wVirtualScanCode ||	/* Right shift */
   		56 == kb->wVirtualScanCode ||	/* Alt */
   		58 == kb->wVirtualScanCode ||	/* Caps lock */
   		69 == kb->wVirtualScanCode ||	/* Num lock */
   		70 == kb->wVirtualScanCode ||	/* Scroll lock */
   		76 == kb->wVirtualScanCode ||	/* Numeric 5, no num lock */
 		91 == kb->wVirtualScanCode ||	/* Left Windows */
   		92 == kb->wVirtualScanCode ||	/* Right Windows */
     	93 == kb->wVirtualScanCode))	/* Menu */
	{
		modifier = 1;
	}
	
	return(modifier);
}


/*
 *	Function:		kbhit
 *	Synopsis:		num = kbhit()
 *	Parameters:		None
 *	Returns:		int num - The number of keypresses waiting to be retrieved from the keyboard input stream
 *	Description:	Checks to see if a key has been pressed and returns true if one
 *					has. This function does not take account of repeat counts of
 *					keys, so it will only return true for the first press of a
 *					key (or subsequent presses if the repeats end up split into
 *					multiple presses).
 *
 *					To maintain compatibility with the standard library getch()
 *					this function will strip out all keypresses that getch()
 *					will not respond to (otherwise checking for new keypresses
 *					becomes rather pointless).
 *
 *					You should use this function (and not enhanced_kbhit()) with the
 *					standard getch() function.
 */
int		kbhit(void)
{
	DWORD			num = 0;	/* Number of pending keyboard input events waiting to be retrieved */
	HANDLE			console;	/* Handle of the console to get access to the input stream */
	INPUT_RECORD	ir;			/* Structure to contain the next input event on the console */

	console = GetStdHandle(STD_INPUT_HANDLE);

	PeekConsoleInput(console,  &ir, 1, &num);
	while(num && (ir.EventType != KEY_EVENT || !ir.Event.KeyEvent.bKeyDown || is_modifier(&ir.Event.KeyEvent)))
	{
		ReadConsoleInput(console,  &ir, 1, &num);
		PeekConsoleInput(console,  &ir, 1, &num);
	}
	
	return(num);
}


/* Used by the kbhit() and getvch() functions so that we can trap and produce
 * any kind of response, including being able to handle multiple repeat counts
 * and unknown extended keys
 */
static INPUT_RECORD	enhanced_ir;


/*
 *	Function:		enhanced_kbhit
 *	Synopsis:		num = enhanced_kbhit()
 *	Parameters:		None
 *	Returns:		int num - The number of keypresses waiting to be retrieved from the keyboard input stream
 *	Description:	Checks to see if a key has been pressed and returns true if one
 *					has. This function does not take account of repeat counts of
 *					keys, so it will only return true for the first press of a
 *					key (or subsequent presses if the repeats end up split into
 *					multiple presses).
 *
 *					You should only use this function with the getvch() function.
 *
 *					This function is not thread safe.
 */
int		enhanced_kbhit(void)
{
	DWORD			num = 0;
	HANDLE			console;

	if(KEY_EVENT == enhanced_ir.EventType && 0 != enhanced_ir.Event.KeyEvent.bKeyDown && enhanced_ir.Event.KeyEvent.wRepeatCount > 0)
	{
		num = 1;
	}
	else
	{
		console = GetStdHandle(STD_INPUT_HANDLE);
	
		PeekConsoleInput(console,  &enhanced_ir, 1, &num);
		while(num && (enhanced_ir.EventType != KEY_EVENT || !enhanced_ir.Event.KeyEvent.bKeyDown || is_modifier(&enhanced_ir.Event.KeyEvent)))
		{
			ReadConsoleInput(console,  &enhanced_ir, 1, &num);
			PeekConsoleInput(console,  &enhanced_ir, 1, &num);
		}
	}
	return(num);
}


/*
 *	Function:		getvch
 *	Synopsis:		keypress = getvch()
 *	Parameters:		None
 *	Returns:		An encoded keycode which means you can get the ASCII, modifiers and extended keys in a single function call
 *	Description:	Similar to the standard getch() function this one will wait
 *					for input from a console
 *
 *					THIS FUNCTION IS NOT THREAD SAFE!
 *
 *					Do not use this function yet. Eventually it will be coded to
 *					be similar to the real getch() but be capable of recognising
 *					and returning some extended keys that getch() doesn't (for
 *					example the Windows keys).
 *
 *					Still need to strip out control keys only and the problem under
 *					Win98 where the cursor keys without numlock turned on
 *					produce a left shift key action (but left shift key is not
 *					in the control state) after the real keypress.
 */
static int		getvch(void)
{
	int		keypress = -1;

	DWORD				num = 0;
	HANDLE				console;
	
	console = GetStdHandle(STD_INPUT_HANDLE);
	
	if(KEY_EVENT != enhanced_ir.EventType || 0 == enhanced_ir.Event.KeyEvent.bKeyDown || enhanced_ir.Event.KeyEvent.wRepeatCount == 0)
	{
		do
		{
			if(!ReadConsoleInput(console,  &enhanced_ir, 1, &num))
			{
				num = 0;
			}
			else
			{
				if(!enhanced_ir.Event.KeyEvent.bKeyDown || enhanced_ir.EventType != KEY_EVENT)
				{
					num = 0;
				}
			}	
		} while(num == 0);
	}
	
	/* Decode the ASCII code or the scan code into the keypress */
	if((enhanced_ir.Event.KeyEvent.dwControlKeyState & ENHANCED_KEY) != 0)
	{
		keypress = enhanced_ir.Event.KeyEvent.wVirtualScanCode << 8 | 224;
	}
	else if(0 == enhanced_ir.Event.KeyEvent.uChar.AsciiChar)
	{
		keypress = enhanced_ir.Event.KeyEvent.wVirtualScanCode << 8;
	}
	else
	{
		keypress = enhanced_ir.Event.KeyEvent.uChar.AsciiChar;
	}

	/* Reduce repeat counter for next time we call this function. This should
 	 * only happen when we have a valid keyboard event, which is when the
     * function will exit.
     */	
	enhanced_ir.Event.KeyEvent.wRepeatCount--;
	
	return(keypress);	
}


/*
 *	Function:		getline
 *	Synopsis:		line_length = getline(line_buffer, buffer_length)
 *	Parameters:		char *line_buffer - Address of the start of a block of memory to be used as the buffer into which the line will be stored
 *					int buffer_length - The length of the buffer which should include the character for the null terminator
 *	Returns:		int line_length - The number of characters stored in the buffer, not including the null terminator.
 *	Description:	Reads a line of characters from the keyboard and stores it in
 *					the specified buffer. The line will be terminated when the user
 *					enters a newline character or the buffer is full. The buffer
 *					will be null terminated (so a full buffer is actually length-1
 *					characters) and the newline will be removed (if that caused
 *					the termination).
 */
int	getline(char *line_buffer, int buffer_length)
{
	int		c;				/* Most recently entered character from the keyboard */
	int		line_length;	/* Current length of line being input */
	
	for(line_length = 0; line_length < buffer_length-1 && (c=getchar())!=EOF && c!='\n'; line_length++)
	{
		if('\r' != c && '\n' != c)
		{
			line_buffer[line_length] = c;
		}
	}
	
	line_buffer[line_length] = '\0';
	return(line_length);
}


/*
 *	Function:		
 *	Synopsis:		
 *	Parameters:		
 *	Returns:		
 *	Description:	
 */

