/*
 *	Filename:		debug.c
 *	Description:	Code to implement the debugging functions. Currently there
 *					are two functions kprintf() to display a debug string and
 *					kmodulename() to get (and keep a copy of) the executable's name.
 *					It is possible to control how the information is displayed
 *					by setting the value of the DEBUG_PRINTF_OUTPUT symbol to be either:
 *						0 - display using the Win32 OutputDebugString() function (default if DEBUG_PRINTF_OUTPUT is not defined)
 *						1 - display using the C standard printf() function
 *					If you use the value 0 then you will need to have some debug
 *					tool which captures and displays the output of the OutputDebugString()
 *					function, such as DebugView from the www.sysinternals.com website.
 *	Example:
 *	
 *		d2(kprintf("%s - " __FILE__ "/%s/%d: \n", kmodulename(), FUNCNAME, __LINE__));
 *
 *	To do:			
 */

#include "debug.h"

#if DEBUG>0

#include <windows.h>

#include <stdio.h>
#include <stdarg.h>

#ifndef DEBUG_PRINTF_OUTPUT
#define DEBUG_PRINTF_OUTPUT	0
#endif


/*
 *	Function:		kprintf
 *	Synopsis:		characters = kprintf(format, ...)
 *	Parameters:		const char *format	- Address of a string containing the message and formatting to be printed. The format and arguments (passed as the varargs parameter) are the same as the C standard library printf().
 *	Returns:		int characters - The number of characters sent to the debug output
 *	Description:	Converts and formats a string and any associated varargs
 *					and sends it to the debug output. By default the debug output
 *					is the Win32 OutputDebugString() destination, but this can
 *					be changed to stdout using the DEBUG_PRINTF_OUTPUT compiler
 *					conditional (see unit description).
 *
 *					Typical use might be like this:
 *
 *	d2(kprintf("%s - " __FILE__ "/%s/%d: \n", kmodulename(), FUNCNAME, __LINE__));
 */
int kprintf(const char *format, ...)
{
	#define DEBUG_BUFFER_LENGTH		256

	int 		characters;							/* Number of characters printed */
	va_list 	ap;									/* For accessing the crap the user gave us */
	char		buffer[DEBUG_BUFFER_LENGTH];		/* Woah! big buffer for the resulting debug string */

	va_start(ap, format);						/* Get the users format string */
	#if DEBUG_PRINTF_OUTPUT==0
	characters = vsprintf(buffer, format, ap);	/* Format the users string and data into a string */
	buffer[DEBUG_BUFFER_LENGTH-1] = '\0';		/* Just in case */
	OutputDebugString(buffer);					/* Spit out the debug string */
	#else
	characters = vprintf(format, ap);
	#endif
	va_end(ap);								/* Finished with the users format string */
	
	return(characters);
}


/*
 *	Function:		kmodulename
 *	Synopsis:		module_name = kmodulename()
 *	Parameters:		None
 *	Returns:		Address of a string in which the module name is stored
 *	Description:	Finds out the name of the current module's (i.e. the program
 *					which is running) executable and returns it as a string without
 *					the path or extension. This can be useful if you want the
 *					program to be able to refer to itself (for example in debug
 *					outputs so you can filter for only the program you are interested in).
 *
 *					Should probably be moved to a more general unit so that it is
 *					available separately from the debug functions.
 */
const char *kmodulename(void)
{
	#define MODULE_NAME_LENGTH	512

	static char	module_name[MODULE_NAME_LENGTH] = {0};	/* Buffer to get and parse program name for inclusion into debug string. Defined as static so that we don't need to keep calculating the exe name every debug string. */
	
	/* Find the executable name and remove any path and extensions from it */
	if('\0' == module_name[0])
	{
		if(GetModuleFileName(NULL, module_name, MODULE_NAME_LENGTH))
		{
			if(strrchr(module_name, '.'))
			{
				*(strrchr(module_name, '.')) = '\0';
			}
			
			if(strrchr(module_name, '\\'))
			{
				strcpy(module_name, strrchr(module_name, '\\')+1);
			}
		}
		else
		{
			strcpy(module_name, "Untitled");
		}
	}
	
	return(&module_name[0]);
}


#endif /* DEBUG>0 */


